function [S_load, SoC_trajectory, charging_periods_week] = GenerateEVChargingStrategy(ev_data, soc_constraints, ev_charging_periods)
    %% This function generates power charging and soc trajectories based on conventional chargning conditions

    n_week = 24*7;
    
    battery_capacity          = ev_data.battery_capacity;                      % Battery capacity in kWh
    eta                       = ev_data.efficiency;                            % Battery efficiency
    charging_power            = ev_data.charging_power;                      % Hur många kW laddas per timme

    
    fully_charged_soc         = soc_constraints.soc_full;                       % What level of SOC when fully charged

    
    charging_periods_week     = reshape(ev_charging_periods.hours.', n_week,1);  % binary matrix [7*24] is 1 during hours 
                                                                               % when vehicle is connected to the grid 
    avg_trip_consumption      = ev_charging_periods.avg_trip_consumption;      % How much kWh are consumed in average per trip                                    % each day of the week
    %% Generate SOC trajectory for one week
    
    
    charging_periods_week = repmat(charging_periods_week, 2,1);
    S_load_week = zeros(2*n_week,1);
    soc_week    = zeros(2*n_week,1);

    % Find first hour after first charging
    idx_start = find([charging_periods_week; 1] == 0 & [0;charging_periods_week] == 1,1);
    % compute initial SOC
    soc_week(idx_start) = fully_charged_soc;
    for hour_idx = idx_start:numel(soc_week)-1
        % If EV vailable for charging
        if charging_periods_week(hour_idx)
            charging = charging_power;
            if soc_week(hour_idx) + charging_power/battery_capacity*eta > fully_charged_soc
                charging = (fully_charged_soc - soc_week(hour_idx))*battery_capacity/eta;
            end
            soc_week(hour_idx+1) = soc_week(hour_idx) + charging/battery_capacity*eta;
            S_load_week(hour_idx) = charging;
            
        elseif ~charging_periods_week(hour_idx)
            if ~charging_periods_week(hour_idx+1)
                soc_week(hour_idx+1) = soc_week(hour_idx);
            else
                soc_week(hour_idx+1) = max(0, soc_week(hour_idx) - avg_trip_consumption);
            end
        end
    end
    
    S_load = [S_load_week(n_week+1:n_week+idx_start-1);S_load_week(idx_start:n_week)];
    SoC_trajectory = [soc_week(n_week+1:n_week+idx_start-1);soc_week(idx_start:n_week)];
end