// Compile script in Matlab using:
// mex -v CXXOPTIMFLAGS="-O3 -fwrapv -ffast-math -march=native" gridLargeTSFS14.cc

#include <iostream>
#include <vector>
#include <algorithm>
#include <cmath>
#include <complex>

/* MEX */
#include "mex.h"

/* DEFINES */
#define SQRT3           1.732050807568877
#define SQRT3inv        0.577350269189626

#define NNODES          35

#define MAX_ITER        100
#define CONV_EPS        1.000000e-15

void CCsolveFBSM( std::vector< std::complex<double> > &Z_ser,
                  std::vector< std::complex<double> > &Y_shu,
                  std::vector< std::complex<double> > &S_in,
                  std::vector< std::complex<double> > &U_in,
                  std::vector< std::complex<double> > &S_out,
                  std::vector< std::complex<double> > &U_out,
                  std::vector< std::complex<double> > &I_out) {

	// Edge Current
	std::vector<std::complex<double>> I(34, 0);
	std::vector<std::complex<double>> Iold(34, 0);
	// Node Power
	std::vector<std::complex<double>> S(35, 0);
	std::vector<std::complex<double>> Sold(35, 0);
	// Node Voltage
	std::vector<std::complex<double>> U(35, 0);
	std::vector<std::complex<double>> Uold(35, 0);

	// S loss
	std::vector<std::complex<double>> S_loss(34, 0);

	//Initiate variables
	S = S_in;
	U = U_in;

	for (int iter = 0; iter < MAX_ITER; iter++ ) {
		// store previous values
		Iold = I;
		Uold = U;
		Sold = S;

		//Reset power
		S = S_in;

		// update power and currents
		I[33] = std::conj(SQRT3inv*S[28]/U[28]);
		S_loss[33] = 3.0*I[33]*std::conj(I[33])*Z_ser[33];
		S[19] = S[19] + S[28] + S_loss[33];

		I[32] = std::conj(SQRT3inv*S[27]/U[27]);
		S_loss[32] = 3.0*I[32]*std::conj(I[32])*Z_ser[32];
		S[19] = S[19] + S[27] + S_loss[32];

		I[31] = std::conj(SQRT3inv*S[26]/U[26]);
		S_loss[31] = 3.0*I[31]*std::conj(I[31])*Z_ser[31];
		S[19] = S[19] + S[26] + S_loss[31];

		I[30] = std::conj(SQRT3inv*S[25]/U[25]);
		S_loss[30] = 3.0*I[30]*std::conj(I[30])*Z_ser[30];
		S[19] = S[19] + S[25] + S_loss[30];

		I[29] = std::conj(SQRT3inv*S[24]/U[24]);
		S_loss[29] = 3.0*I[29]*std::conj(I[29])*Z_ser[29];
		S[19] = S[19] + S[24] + S_loss[29];

		I[28] = std::conj(SQRT3inv*S[23]/U[23]);
		S_loss[28] = 3.0*I[28]*std::conj(I[28])*Z_ser[28];
		S[19] = S[19] + S[23] + S_loss[28];

		I[27] = std::conj(SQRT3inv*S[22]/U[22]);
		S_loss[27] = 3.0*I[27]*std::conj(I[27])*Z_ser[27];
		S[19] = S[19] + S[22] + S_loss[27];

		I[26] = std::conj(SQRT3inv*S[21]/U[21]);
		S_loss[26] = 3.0*I[26]*std::conj(I[26])*Z_ser[26];
		S[19] = S[19] + S[21] + S_loss[26];

		I[25] = std::conj(SQRT3inv*S[20]/U[20]);
		S_loss[25] = 3.0*I[25]*std::conj(I[25])*Z_ser[25];
		S[19] = S[19] + S[20] + S_loss[25];

		I[24] = std::conj(SQRT3inv*S[34]/U[34]);
		S_loss[24] = 3.0*I[24]*std::conj(I[24])*Z_ser[24];
		S[18] = S[18] + S[34] + S_loss[24];

		I[23] = std::conj(SQRT3inv*S[33]/U[33]);
		S_loss[23] = 3.0*I[23]*std::conj(I[23])*Z_ser[23];
		S[18] = S[18] + S[33] + S_loss[23];

		I[22] = std::conj(SQRT3inv*S[32]/U[32]);
		S_loss[22] = 3.0*I[22]*std::conj(I[22])*Z_ser[22];
		S[18] = S[18] + S[32] + S_loss[22];

		I[21] = std::conj(SQRT3inv*S[31]/U[31]);
		S_loss[21] = 3.0*I[21]*std::conj(I[21])*Z_ser[21];
		S[18] = S[18] + S[31] + S_loss[21];

		I[20] = std::conj(SQRT3inv*S[30]/U[30]);
		S_loss[20] = 3.0*I[20]*std::conj(I[20])*Z_ser[20];
		S[18] = S[18] + S[30] + S_loss[20];

		I[19] = std::conj(SQRT3inv*S[29]/U[29]);
		S_loss[19] = 3.0*I[19]*std::conj(I[19])*Z_ser[19];
		S[18] = S[18] + S[29] + S_loss[19];

		I[18] = std::conj(SQRT3inv*S[19]/U[19]);
		S_loss[18] = 3.0*I[18]*std::conj(I[18])*Z_ser[18];
		S[18] = S[18] + S[19] + S_loss[18];

		I[17] = std::conj(SQRT3inv*S[13]/U[13]);
		S_loss[17] = 3.0*I[17]*std::conj(I[17])*Z_ser[17];
		S[4] = S[4] + S[13] + S_loss[17];

		I[16] = std::conj(SQRT3inv*S[12]/U[12]);
		S_loss[16] = 3.0*I[16]*std::conj(I[16])*Z_ser[16];
		S[4] = S[4] + S[12] + S_loss[16];

		I[15] = std::conj(SQRT3inv*S[11]/U[11]);
		S_loss[15] = 3.0*I[15]*std::conj(I[15])*Z_ser[15];
		S[4] = S[4] + S[11] + S_loss[15];

		I[14] = std::conj(SQRT3inv*S[10]/U[10]);
		S_loss[14] = 3.0*I[14]*std::conj(I[14])*Z_ser[14];
		S[4] = S[4] + S[10] + S_loss[14];

		I[13] = std::conj(SQRT3inv*S[9]/U[9]);
		S_loss[13] = 3.0*I[13]*std::conj(I[13])*Z_ser[13];
		S[4] = S[4] + S[9] + S_loss[13];

		I[12] = std::conj(SQRT3inv*S[8]/U[8]);
		S_loss[12] = 3.0*I[12]*std::conj(I[12])*Z_ser[12];
		S[4] = S[4] + S[8] + S_loss[12];

		I[11] = std::conj(SQRT3inv*S[7]/U[7]);
		S_loss[11] = 3.0*I[11]*std::conj(I[11])*Z_ser[11];
		S[4] = S[4] + S[7] + S_loss[11];

		I[10] = std::conj(SQRT3inv*S[6]/U[6]);
		S_loss[10] = 3.0*I[10]*std::conj(I[10])*Z_ser[10];
		S[4] = S[4] + S[6] + S_loss[10];

		I[9] = std::conj(SQRT3inv*S[5]/U[5]);
		S_loss[9] = 3.0*I[9]*std::conj(I[9])*Z_ser[9];
		S[4] = S[4] + S[5] + S_loss[9];

		I[8] = std::conj(SQRT3inv*S[17]/U[17]);
		S_loss[8] = 3.0*I[8]*std::conj(I[8])*Z_ser[8];
		S[3] = S[3] + S[17] + S_loss[8];

		I[7] = std::conj(SQRT3inv*S[16]/U[16]);
		S_loss[7] = 3.0*I[7]*std::conj(I[7])*Z_ser[7];
		S[3] = S[3] + S[16] + S_loss[7];

		I[6] = std::conj(SQRT3inv*S[15]/U[15]);
		S_loss[6] = 3.0*I[6]*std::conj(I[6])*Z_ser[6];
		S[3] = S[3] + S[15] + S_loss[6];

		I[5] = std::conj(SQRT3inv*S[14]/U[14]);
		S_loss[5] = 3.0*I[5]*std::conj(I[5])*Z_ser[5];
		S[3] = S[3] + S[14] + S_loss[5];

		I[4] = std::conj(SQRT3inv*S[4]/U[4]);
		S_loss[4] = 3.0*I[4]*std::conj(I[4])*Z_ser[4];
		S[3] = S[3] + S[4] + S_loss[4];

		I[3] = std::conj(SQRT3inv*S[18]/U[18]);
		S_loss[3] = 3.0*I[3]*std::conj(I[3])*Z_ser[3];
		S[1] = S[1] + S[18] + S_loss[3];

		I[2] = std::conj(SQRT3inv*S[3]/U[3]);
		S_loss[2] = 3.0*I[2]*std::conj(I[2])*Z_ser[2];
		S[1] = S[1] + S[3] + S_loss[2];

		I[1] = std::conj(SQRT3inv*S[2]/U[2]);
		S_loss[1] = 3.0*I[1]*std::conj(I[1])*Z_ser[1];
		S[1] = S[1] + S[2] + S_loss[1];

		I[0] = std::conj(SQRT3inv*S[1]/U[1]);
		S_loss[0] = 3.0*I[0]*std::conj(I[0])*Z_ser[0];
		S[0] = S[0] + S[1] + S_loss[0];

		// update voltages
		U[1] = U[0] - SQRT3*I[0]*Z_ser[0];
		U[2] = U[1] - SQRT3*I[1]*Z_ser[1];
		U[3] = U[1] - SQRT3*I[2]*Z_ser[2];
		U[18] = U[1] - SQRT3*I[3]*Z_ser[3];
		U[4] = U[3] - SQRT3*I[4]*Z_ser[4];
		U[14] = U[3] - SQRT3*I[5]*Z_ser[5];
		U[15] = U[3] - SQRT3*I[6]*Z_ser[6];
		U[16] = U[3] - SQRT3*I[7]*Z_ser[7];
		U[17] = U[3] - SQRT3*I[8]*Z_ser[8];
		U[5] = U[4] - SQRT3*I[9]*Z_ser[9];
		U[6] = U[4] - SQRT3*I[10]*Z_ser[10];
		U[7] = U[4] - SQRT3*I[11]*Z_ser[11];
		U[8] = U[4] - SQRT3*I[12]*Z_ser[12];
		U[9] = U[4] - SQRT3*I[13]*Z_ser[13];
		U[10] = U[4] - SQRT3*I[14]*Z_ser[14];
		U[11] = U[4] - SQRT3*I[15]*Z_ser[15];
		U[12] = U[4] - SQRT3*I[16]*Z_ser[16];
		U[13] = U[4] - SQRT3*I[17]*Z_ser[17];
		U[19] = U[18] - SQRT3*I[18]*Z_ser[18];
		U[29] = U[18] - SQRT3*I[19]*Z_ser[19];
		U[30] = U[18] - SQRT3*I[20]*Z_ser[20];
		U[31] = U[18] - SQRT3*I[21]*Z_ser[21];
		U[32] = U[18] - SQRT3*I[22]*Z_ser[22];
		U[33] = U[18] - SQRT3*I[23]*Z_ser[23];
		U[34] = U[18] - SQRT3*I[24]*Z_ser[24];
		U[20] = U[19] - SQRT3*I[25]*Z_ser[25];
		U[21] = U[19] - SQRT3*I[26]*Z_ser[26];
		U[22] = U[19] - SQRT3*I[27]*Z_ser[27];
		U[23] = U[19] - SQRT3*I[28]*Z_ser[28];
		U[24] = U[19] - SQRT3*I[29]*Z_ser[29];
		U[25] = U[19] - SQRT3*I[30]*Z_ser[30];
		U[26] = U[19] - SQRT3*I[31]*Z_ser[31];
		U[27] = U[19] - SQRT3*I[32]*Z_ser[32];
		U[28] = U[19] - SQRT3*I[33]*Z_ser[33];
		// check if converged
		if(std::abs(S[0] - Sold[0]) > CONV_EPS) continue;
		if(std::abs(S[1] - Sold[1]) > CONV_EPS) continue;
		if(std::abs(S[2] - Sold[2]) > CONV_EPS) continue;
		if(std::abs(S[3] - Sold[3]) > CONV_EPS) continue;
		if(std::abs(S[4] - Sold[4]) > CONV_EPS) continue;
		if(std::abs(S[5] - Sold[5]) > CONV_EPS) continue;
		if(std::abs(S[6] - Sold[6]) > CONV_EPS) continue;
		if(std::abs(S[7] - Sold[7]) > CONV_EPS) continue;
		if(std::abs(S[8] - Sold[8]) > CONV_EPS) continue;
		if(std::abs(S[9] - Sold[9]) > CONV_EPS) continue;
		if(std::abs(S[10] - Sold[10]) > CONV_EPS) continue;
		if(std::abs(S[11] - Sold[11]) > CONV_EPS) continue;
		if(std::abs(S[12] - Sold[12]) > CONV_EPS) continue;
		if(std::abs(S[13] - Sold[13]) > CONV_EPS) continue;
		if(std::abs(S[14] - Sold[14]) > CONV_EPS) continue;
		if(std::abs(S[15] - Sold[15]) > CONV_EPS) continue;
		if(std::abs(S[16] - Sold[16]) > CONV_EPS) continue;
		if(std::abs(S[17] - Sold[17]) > CONV_EPS) continue;
		if(std::abs(S[18] - Sold[18]) > CONV_EPS) continue;
		if(std::abs(S[19] - Sold[19]) > CONV_EPS) continue;
		if(std::abs(S[20] - Sold[20]) > CONV_EPS) continue;
		if(std::abs(S[21] - Sold[21]) > CONV_EPS) continue;
		if(std::abs(S[22] - Sold[22]) > CONV_EPS) continue;
		if(std::abs(S[23] - Sold[23]) > CONV_EPS) continue;
		if(std::abs(S[24] - Sold[24]) > CONV_EPS) continue;
		if(std::abs(S[25] - Sold[25]) > CONV_EPS) continue;
		if(std::abs(S[26] - Sold[26]) > CONV_EPS) continue;
		if(std::abs(S[27] - Sold[27]) > CONV_EPS) continue;
		if(std::abs(S[28] - Sold[28]) > CONV_EPS) continue;
		if(std::abs(S[29] - Sold[29]) > CONV_EPS) continue;
		if(std::abs(S[30] - Sold[30]) > CONV_EPS) continue;
		if(std::abs(S[31] - Sold[31]) > CONV_EPS) continue;
		if(std::abs(S[32] - Sold[32]) > CONV_EPS) continue;
		if(std::abs(S[33] - Sold[33]) > CONV_EPS) continue;
		if(std::abs(S[34] - Sold[34]) > CONV_EPS) continue;
		if(std::abs(U[0] - Uold[0]) > CONV_EPS) continue;
		if(std::abs(U[1] - Uold[1]) > CONV_EPS) continue;
		if(std::abs(U[2] - Uold[2]) > CONV_EPS) continue;
		if(std::abs(U[3] - Uold[3]) > CONV_EPS) continue;
		if(std::abs(U[4] - Uold[4]) > CONV_EPS) continue;
		if(std::abs(U[5] - Uold[5]) > CONV_EPS) continue;
		if(std::abs(U[6] - Uold[6]) > CONV_EPS) continue;
		if(std::abs(U[7] - Uold[7]) > CONV_EPS) continue;
		if(std::abs(U[8] - Uold[8]) > CONV_EPS) continue;
		if(std::abs(U[9] - Uold[9]) > CONV_EPS) continue;
		if(std::abs(U[10] - Uold[10]) > CONV_EPS) continue;
		if(std::abs(U[11] - Uold[11]) > CONV_EPS) continue;
		if(std::abs(U[12] - Uold[12]) > CONV_EPS) continue;
		if(std::abs(U[13] - Uold[13]) > CONV_EPS) continue;
		if(std::abs(U[14] - Uold[14]) > CONV_EPS) continue;
		if(std::abs(U[15] - Uold[15]) > CONV_EPS) continue;
		if(std::abs(U[16] - Uold[16]) > CONV_EPS) continue;
		if(std::abs(U[17] - Uold[17]) > CONV_EPS) continue;
		if(std::abs(U[18] - Uold[18]) > CONV_EPS) continue;
		if(std::abs(U[19] - Uold[19]) > CONV_EPS) continue;
		if(std::abs(U[20] - Uold[20]) > CONV_EPS) continue;
		if(std::abs(U[21] - Uold[21]) > CONV_EPS) continue;
		if(std::abs(U[22] - Uold[22]) > CONV_EPS) continue;
		if(std::abs(U[23] - Uold[23]) > CONV_EPS) continue;
		if(std::abs(U[24] - Uold[24]) > CONV_EPS) continue;
		if(std::abs(U[25] - Uold[25]) > CONV_EPS) continue;
		if(std::abs(U[26] - Uold[26]) > CONV_EPS) continue;
		if(std::abs(U[27] - Uold[27]) > CONV_EPS) continue;
		if(std::abs(U[28] - Uold[28]) > CONV_EPS) continue;
		if(std::abs(U[29] - Uold[29]) > CONV_EPS) continue;
		if(std::abs(U[30] - Uold[30]) > CONV_EPS) continue;
		if(std::abs(U[31] - Uold[31]) > CONV_EPS) continue;
		if(std::abs(U[32] - Uold[32]) > CONV_EPS) continue;
		if(std::abs(U[33] - Uold[33]) > CONV_EPS) continue;
		if(std::abs(U[34] - Uold[34]) > CONV_EPS) continue;
		if(std::abs(I[0] - Iold[0]) > CONV_EPS) continue;
		if(std::abs(I[1] - Iold[1]) > CONV_EPS) continue;
		if(std::abs(I[2] - Iold[2]) > CONV_EPS) continue;
		if(std::abs(I[3] - Iold[3]) > CONV_EPS) continue;
		if(std::abs(I[4] - Iold[4]) > CONV_EPS) continue;
		if(std::abs(I[5] - Iold[5]) > CONV_EPS) continue;
		if(std::abs(I[6] - Iold[6]) > CONV_EPS) continue;
		if(std::abs(I[7] - Iold[7]) > CONV_EPS) continue;
		if(std::abs(I[8] - Iold[8]) > CONV_EPS) continue;
		if(std::abs(I[9] - Iold[9]) > CONV_EPS) continue;
		if(std::abs(I[10] - Iold[10]) > CONV_EPS) continue;
		if(std::abs(I[11] - Iold[11]) > CONV_EPS) continue;
		if(std::abs(I[12] - Iold[12]) > CONV_EPS) continue;
		if(std::abs(I[13] - Iold[13]) > CONV_EPS) continue;
		if(std::abs(I[14] - Iold[14]) > CONV_EPS) continue;
		if(std::abs(I[15] - Iold[15]) > CONV_EPS) continue;
		if(std::abs(I[16] - Iold[16]) > CONV_EPS) continue;
		if(std::abs(I[17] - Iold[17]) > CONV_EPS) continue;
		if(std::abs(I[18] - Iold[18]) > CONV_EPS) continue;
		if(std::abs(I[19] - Iold[19]) > CONV_EPS) continue;
		if(std::abs(I[20] - Iold[20]) > CONV_EPS) continue;
		if(std::abs(I[21] - Iold[21]) > CONV_EPS) continue;
		if(std::abs(I[22] - Iold[22]) > CONV_EPS) continue;
		if(std::abs(I[23] - Iold[23]) > CONV_EPS) continue;
		if(std::abs(I[24] - Iold[24]) > CONV_EPS) continue;
		if(std::abs(I[25] - Iold[25]) > CONV_EPS) continue;
		if(std::abs(I[26] - Iold[26]) > CONV_EPS) continue;
		if(std::abs(I[27] - Iold[27]) > CONV_EPS) continue;
		if(std::abs(I[28] - Iold[28]) > CONV_EPS) continue;
		if(std::abs(I[29] - Iold[29]) > CONV_EPS) continue;
		if(std::abs(I[30] - Iold[30]) > CONV_EPS) continue;
		if(std::abs(I[31] - Iold[31]) > CONV_EPS) continue;
		if(std::abs(I[32] - Iold[32]) > CONV_EPS) continue;
		if(std::abs(I[33] - Iold[33]) > CONV_EPS) continue;
		break;
	}

	//return results
	S_out = S;
	U_out = U;
	I_out = I;
}

// Main MEX-function
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	// Pointers to MATLAB inputs (M_. denotes a MATLAB input)
	double *M_Z_serR = mxGetPr(prhs[0]);
	double *M_Z_serI = mxGetPi(prhs[0]);
	double *M_Y_shuR = mxGetPr(prhs[1]);
	double *M_Y_shuI = mxGetPi(prhs[1]);
	double *M_S_inR = mxGetPr(prhs[2]);
	double *M_S_inI = mxGetPi(prhs[2]);
	double *M_U_inR = mxGetPr(prhs[3]);
	double *M_U_inI = mxGetPi(prhs[3]);

	int nTime = mxGetScalar(prhs[4]);

	// Z ser
	std::vector<std::complex<double>> Z_ser(34, 0);
	// Y shu
	std::vector<std::complex<double>> Y_shu(34, 0);

	for ( int iEdge = 0; iEdge < 34; iEdge++ ) {
		Z_ser[iEdge].real(M_Z_serR[iEdge]);
		Z_ser[iEdge].imag(M_Z_serI[iEdge]);
		Y_shu[iEdge].real(M_Y_shuR[iEdge]);
		Y_shu[iEdge].real(M_Y_shuI[iEdge]);
	}
	// Convert S_in and U_in from vectors to matrices
	std::vector< std::vector<std::complex<double>> > CS_in(nTime, std::vector<std::complex<double>> (35,0));
	std::vector< std::vector<std::complex<double>> > CU_in(nTime, std::vector<std::complex<double>> (35,0));

	for ( int iTime = 0; iTime < nTime; iTime++ ) {
		for ( int iBus = 0; iBus < 35; iBus++ ) {
			CS_in[iTime][iBus].real(M_S_inR[NNODES*iTime + iBus]);
			CS_in[iTime][iBus].imag(M_S_inI[NNODES*iTime + iBus]);
			CU_in[iTime][iBus].real(M_U_inR[NNODES*iTime + iBus]);
			CU_in[iTime][iBus].imag(M_U_inI[NNODES*iTime + iBus]);
		}
	}
	// Store results in CS_out, CU_out, CI_out
	std::vector< std::vector<std::complex<double>> > CS_out(nTime, std::vector<std::complex<double>> (35,0));
	std::vector< std::vector<std::complex<double>> > CU_out(nTime, std::vector<std::complex<double>> (35,0));
	std::vector< std::vector<std::complex<double>> > CI_out(nTime, std::vector<std::complex<double>> (35,0));

	// Run Forward-Backward Sweep for each time step
	for ( int iTime = 0; iTime < nTime; iTime++ ) {
		// Call the C subroutine.
		CCsolveFBSM(Z_ser, Y_shu, CS_in[iTime], CU_in[iTime], CS_out[iTime], CU_out[iTime], CI_out[iTime]);
	}

	// S_out
	plhs[0] = mxCreateDoubleMatrix(35, nTime, mxCOMPLEX);
	double *M_S_outR = mxGetPr(plhs[0]);
	double *M_S_outI = mxGetPi(plhs[0]);

	// U_out
	plhs[1] = mxCreateDoubleMatrix(35, nTime, mxCOMPLEX);
	double *M_U_outR = mxGetPr(plhs[1]);
	double *M_U_outI = mxGetPi(plhs[1]);

	// I_out
	plhs[2] = mxCreateDoubleMatrix(34, nTime, mxCOMPLEX);
	double *M_I_outR = mxGetPr(plhs[2]);
	double *M_I_outI = mxGetPi(plhs[2]);

	// Return results to Matlab
	for ( int iTime = 0; iTime < nTime; iTime++ ) {
		for ( int iNode = 0; iNode < 35; iNode++ ) {
			M_S_outR[35*iTime + iNode] = std::real(CS_out[iTime][iNode]);
			M_S_outI[35*iTime + iNode] = std::imag(CS_out[iTime][iNode]);
			M_U_outR[35*iTime + iNode] = std::real(CU_out[iTime][iNode]);
			M_U_outI[35*iTime + iNode] = std::imag(CU_out[iTime][iNode]);
		}
	}
	for ( int iTime = 0; iTime < nTime; iTime++ ) {
		for ( int iEdge = 0; iEdge < 34; iEdge++ ) {
			M_I_outR[34*iTime + iEdge] = std::real(CI_out[iTime][iEdge]);
			M_I_outI[34*iTime + iEdge] = std::imag(CI_out[iTime][iEdge]);
		}
	}
}
